; Aberration analysis of ring images using av. coefficients. June 1, 2022
; uses ztools and getimage. see cubecoef.pro for coef. definition.
;-----------------------------------------------------------------
; Image centering
pro recenter, img, dx, dy

common aber1, nx,xx,yy,nsect,xsect,ysect, m,nz,maskmat

nx2 = nx*nx
  t = maskmat[0:2*nsect-1,*] # reform(img,nx2)
  radii = t[0:nsect-1]/t[nsect:2*nsect-1] ; radii, pixels
  dx = total(radii*xsect)/nsect*2.3
  dy = total(radii*ysect)/nsect*2.3

; FFT shift to re-center
  arg = 2.*!pi*(dx*xx + dy*yy)/nx
  img = float( fft(fft(img)*shift(complex(cos(arg), sin(arg)), nx/2, nx/2), /inverse)) 

end  
;-----------------------------------------------------------------
pro abercoef, par  ; get sparse interaction matrix and aberration coefficients

common aber1, nx,xx,yy,nsect,xsect,ysect, m,nz,maskmat

nz = 10  ; max Zernike term to ashtray inclusive
m = 3  ; max angular term
nsect = 8  ; number of sectors

getimage, par, img0, rad=rad

radpix = rad/par.pixel  ; ring radius in pixels
nx = (size(img0))[1]

tvscl, congrid(img0,256,256)

; Prepare the matrix - see cubecoef.pro
; Vectors of radius and coordinates
xx = (findgen(nx) - nx/2) # replicate(1,nx)
yy =  transpose(xx)
r = sqrt(xx^2 + yy^2)
phi = atan(yy, xx) ; + !pi ; 0-2pi range
phi[0:nx/2,nx/2] = -!pi ; was pi and created empty ray
x1d = findgen(nx)  ; 1-D coordinate for interpolation

; ring width
rwidth = par.lambda/par.d/(1. - par.eps)*2.*206265./par.pixel 
drhopix = 1.5*rwidth ; mask width, pixels
ringmask = (r ge radpix-drhopix) and (r le radpix+drhopix) ; Circular, use only for display

ncoef = 2*nsect + 2*(m+1)
nx2 = nx*nx
maskmat = fltarr(ncoef, nx2)

rwt = fltarr(nx,nx,nsect)  ; for radius calculation
fwt = fltarr(nx,nx,nsect)  ; fluxes in the sectors
sect = 2.*!pi/nsect ; sector with, radian
for j=0,nsect-1 do begin  ; un-masked sectors for initial estimate
   sector = (phi ge (sect*(j- nsect/2))) and (phi lt (sect*(j+1-nsect/2)))
   fwt[*,*,j] = sector*ringmask
   rwt[*,*,j] = sector*r*ringmask 
endfor ; j sectors
; Parameters for ring re-centering
phisect = sect*(findgen(nsect) - nsect/2 +0.5) ; sector angles
xsect = cos(phisect) & ysect = sin(phisect)

; Mask sectors radially, insert into the matrix
for j=0,nsect-1 do begin
  maskmat[j,*] = reform(rwt[*,*,j],nx2)
  maskmat[j+nsect,*] = reform(fwt[*,*,j],nx2)
endfor ;j  

cwt = fltarr(nx,nx,m+1) ; cosine weight
swt = fltarr(nx,nx,m+1) ; sine weight
for j=0,m do begin 
  tmp =  cos(phi*j)*ringmask
  if (j gt 0) then cwt[*,*,j] = tmp  - total(tmp)/nx/nx else cwt[*,*,j] = tmp 
  tmp = sin(phi*j)*ringmask
  swt[*,*,j] = tmp  - total(tmp)/nx/nx
  maskmat[2*nsect+j,*] =  reform(cwt[*,*,j],nx2)
  maskmat[2*nsect+m+1+j,*] = reform(swt[*,*,j],nx2)
endfor ; j angular

print, 'Matrix done. .C to continue with Imat'
;stop

coef = fltarr(ncoef, nz) ; all coefficients
;  xcent = fltarr(nz) & ycent = fltarr(nz) & rad = fltarr(nz)

; loop over Zernike terms
ampl = 0.5 ; test amplitude, radian
for i=0,nz-1 do begin

  getimage, par, img, zn=[i+1], zrad=[ampl]  ; aberrated image
  recenter, img, dx, dy
  print, i+1, dx, dy						 
  c = maskmat # reform(img, nx2)
  c[0:nsect-1] = c[0:nsect-1]/c[nsect:2*nsect-1] ; normalize radius
  coef[*,i] = c
  tvscl, congrid(img,256,256)
endfor

print, 'Coefficients are computed'
;stop

meanrad = total(coef[0:nsect-1,0])/nsect ; unaberrated ring radius
coef[0:nsect-1,*] -= meanrad

; Select only radii and discard m=0 term. nsect+2*m elements
index = [indgen(nsect), indgen(m)+2*nsect+1, indgen(m)+2*nsect+m+2] ; remove fluxes

zindex = indgen(nz-3) + 3  ; 7 Zernike terms without piston and tilts
imat = coef[index,*]/ampl
imat = imat[*,zindex]  ; [14,7] nsect+2*m by nz-3

;s = size(imat)
;tvscl, congrid(imat, s[1]*4, s[2]*20)
;stop

; Fabricate signals
smat = fltarr(7, nsect+2*m)
smat[0,0:7] = 1.              ; focus  
smat[1,0:7] = sin(2*phisect)  ; astig a5
smat[2,0:7] = cos(2*phisect)
smat[3,11] = 1.   ; coma a8 imat[12,2]=0.10
smat[4,8] = 1.   ; coma a7
smat[5,13] = 1.   ; trefoil -0.16
smat[6,10] = 1.   ; trefoil

abercoef = fltarr(7)
for i=0,6 do abercoef[i] = total(smat[i,*]*imat[*,i])
; 8.00701    2.478   2.43896  0.102554  0.098  -0.165 -0.165


s1 = 'abercoef: ['
for i=0,6 do s1 += string(abercoef[i],'(F6.3)')+','
s1 = strmid(s1,0,strlen(s1)-1) + '],  ; for aberration calc.'

;print, 'abercoef: ', abercoef, format='(A,7F8.3)'
print, s1
print, 'Include these coefficients in the parameters!'

;save, fi='imat.idl', imat, index, maskmat, meanrad,smat,abercoef
;print, 'Imat is computed and saved!'
;stop

end
;----------------------------------------------------------------
;------------------------------------------------------------------------------
; retrieve aberrations
; ind=ind:  use particular record, otherwise mean
; optionally returns the aberrated ring image
pro getaber, par, ind=ind,  img=img

;restore, 'imat.idl' ; imat, index, maskmat, meanrad,smat,rcoef

restore, par.datfile ; dat

n = n_elements(dat)
nm = n_elements(dat[0].mcoef)  ; nm=14 

; Construct the S matrix
nsect = par.nsect
sect = 2.*!pi/nsect ; sector with, radian
phisect = sect*(findgen(nsect) - nsect/2 +0.5) ; sector angles

smat = fltarr(7, nm)
smat[0,0:7] = 1.              ; focus  
smat[1,0:7] = sin(2*phisect)  ; astig a5
smat[2,0:7] = cos(2*phisect)
smat[3,11] = 1.   ; coma a8 imat[12,2]=0.10
smat[4,8] = 1.   ; coma a7
smat[5,13] = 1.   ; trefoil -0.16
smat[6,10] = 1.   ; trefoil


if keyword_set(ind) then begin
    print, 'Aberrations for the file '+dat[ind].fname 
    mcoef = dat[ind].mcoef 
endif else   mcoef = total(dat.mcoef, 2)/n 

meanrad = total(mcoef[0:7])/8.

print, 'Ring radius [par, real]: ', par.ringradpix, meanrad, format='(A,2F8.2)'

;restore, par.weightfile 
;rad2 = hslope/par.pdist  ; nominal radius
;print, 'Ring radius [pdist]: ', rad2, format='(A,2F8.2)'


mcoef[0:7] -= par.ringradpix
;mcoef[0:7] -= mean(dat.rad)

; amplitude of terms 4-10 in radian at par.lambda
zrad = (smat # mcoef) /par.abercoef  

; print, 'Zernike amplitudes [4-9], rad: ', zrad, format='(A,7F6.2)'
; for updated parameter  file:
s1 = 'znum: '
for i=0,6 do s1 += string(i+4,'(I2)')+' '

s2 = 'zrad: '
for i=0,6 do s2 += string(zrad[i],'(F5.2)')+' '
; s2 = strmid(s2,0,strlen(s2)-1) + ' ;  ; Zernike ampl, rad'

print, s1
print, s2

; save aberrations in the text file
if (par.aber gt '') then begin 
   get_lun, lun
   openw, lun, par.aber
   printf, lun, s1
   printf, lun, s2
   close, lun
   free_lun, lun
endif

; optionally reconstruct the aberrated ring image
if keyword_set(img) then begin 
  print, 'Computing the aberrated image'
  zn = indgen(9)+2   ; Zernike number
  zrad1 = [-5.*zrad[4], -5.*zrad[3], zrad] 
  getimage, par, img, zn=zn, zrad=zrad1
  tvscl, congrid(img,256,256)
endif

end
;------------------------------------------------------------------------------
;------------------------------------------------------------------------------

