; Based on respmass.pro and respmasspoly.pro, similar to smass.pro
; response of a defocused image angular power spectrum to turbulence
; Apr 15, 2021
; compile ztools if aberrations are specified
;--------------------------------------------------------------
FUNCTION f1, zeta  ; Approximation of temporal filter, zeta=freq*shift 
t = (1.+zeta^2)^(-0.5)
coef = [0., 0.224536, 2.02472, -15.7485, 58.8973, -109.331, 97.6583, -32.7291]
return, poly(t,coef)
END 
;-------------------------------------------------------
function aweight3, z, mmax, d, eps, pdist, wav, sp, drho=drho, pixel=pixel, $
   wtpup=wtpup, zn=zn, zrad=zrad, blur=blur, ufunc=ufunc, ringrad=ringrad
; Input parameters: distance, max.ang.freq,diameter, obscur, wavelength array, sp. array
; if 1-element wavelength - monochromatic
; m=0 weight is for the 8-sector radius var. in lambda/D units
; drho = radial mask width in FWHM units, default=1.5
; optional inputs: pixel size (arcsec), blur during exposure [m]
; output: wt[m,nz] weight in m^(-1/3)
; optional output wtpup is the pupil-plane weight
; optional output Ufunc (Kornilov, 2011) in [m^(-7/3)]
;--------------------------------------------------
;
; --- Hard-coded parameters: good: ngrid=256 ksize=6
NGRID = 256  ; half-size of the simulation grid, pixels 
;NGRID = 512  ; half-size of the simulation grid, pixels 
ksize = 6       ; GRID SIZE/diameter
nsect = 8       ; number of radial sectors

nwav = n_elements(wav)
if (nwav gt 1) then begin ; spectrum is given
 spnorm = sp/total(sp) ; normalized spectrum
 lambda = total(wav*spnorm) ; mean wavelength
 wavstep = (wav[nwav-1] - wav[0])/(nwav-1) ; wavelength grid step
endif else begin  ; central wavelength and bandwidth
 lambda = wav[0]
 spnorm = 1.
 wavstep = 0
endelse 

 
SIZE = D*KSIZE   ; grid size, m; Diameter is 2*ngrid/ksize pixels
asperpix =  206265.*lambda/size        ; pixel in image plane, arcsec
fstep = 1./size    ; pixel size in spatial frequency plane, 1/m
xstep = size/(2*ngrid) ; sampling of the pupil grid
; print, 'Image pixel and  size [arcsec]:  ',asperpix, asperpix*2*ngrid
ringradpix = 0.85*d*(1.+eps)/(4.*pdist)*206265./asperpix
;print, 'Ring radius [pixels]: ', ringradpix

; ----- Prepare the arrays --------------------------------
r = shift(dist(2*ngrid),ngrid,ngrid) ; radial distance from grid center, pixels
r[ngrid,ngrid] = 1.e-3
x = (findgen(2*ngrid) - ngrid) # replicate(1., 2*ngrid) ; X-coordinate, pixels

; ----   Circular aperture with central obscuration
Radpix = ngrid/size*D   ; aperture radius, pixels
inside = where((r le Radpix) and (r ge Radpix*eps))
ninside = n_elements(inside)  ; pupil area in pixels
surface = double(ninside)

ur = fltarr(2*ngrid,2*ngrid)    ; zero array
ur(inside)=1.    ; input amplitude: real part
uampl = complex(ur, ur*0.)

; Propagation filter, positive for intrafocal (negative conjugation) 
; Phase filter: defocus and spherical
a4 = d^2/(lambda*pdist) *(!pi/8/sqrt(3.)) ; Zernike in radians
a11 = -0.1*a4  ; matching spherical

tmp = a11*sqrt(5.)*(6.*(r/Radpix)^4 - 6.*(r/Radpix)^2)
tmp += a4*2.*sqrt(3.)*((r/Radpix)^2 - 0.5)

; Add Zernike aberrations if specified, use ztools. See sim
if keyword_set(zn) then begin
  theta = atan(transpose(x),x)
  zgrid = fltarr(ninside,2)
  zgrid[*,0] = r(inside)/Radpix
  zgrid[*,1] = theta(inside)
  ; -- use Ztools.pro here! ---------------------
  phase = zgrid[*,0]*0. ; empty array
  for k=0,n_elements(ZRAD)-1 do phase += ZRAD[k]*zernike_estim(zn[k], zgrid)
  tmp[inside] += phase
endif

tmp = complex(cos(tmp), sin(tmp))
uampl *= tmp

; ---- Get the image (for normalization) --------------------------------
imh = abs(shift(fft(shift(uampl,ngrid,ngrid),/inverse), ngrid, ngrid))^2
imh = imh/(2.*ngrid)^2 ; total(imh)=surface

; Compute the mask
ringradpix2 =   total(imh*r)/total(imh) ; true ring radius in fine pixels

if keyword_set(ringrad) then begin
;   stop
   return, ringradpix2*asperpix/pixel  ; only radius is needed
endif
   
phi = atan(transpose(x), x)
if not keyword_set(drho) then drho = 1.5  ; radial width in FWHM units 
drhopix = drho*lambda/d/(1. - eps)*2.*206265./asperpix ; drho*(ring half-width), pixels
filtap = (r ge ringradpix2-drhopix) and (r le ringradpix2+drhopix) ; Circular

; Compute the mask for radius fluctuations in two sectors
sectrad = !pi/nsect ; sector half-width in radian
tmp = (phi + !pi) mod !pi
sector = ((tmp ge !pi/2.-sectrad) and (tmp lt !pi/2.+ sectrad))
sector = transpose(sector)
ringmask =  filtap*sector ; 1 inside two opposite sectors
rwt = r * ringmask        ; same with radial gradient
rflux = total(imh*ringmask)
wf0 = total(imh*rwt)/rflux ; piston term
rwt -= wf0*ringmask    ; subtract piston
; pixel = (lambda/D)/ksize radian
rwt = rwt/ksize    ; radius expressed in lambda/D units

;tvscl, rwt ; linear within sectors, zero elsewhere
;stop

;------- loops over m and z ---------------
nz = n_elements(z)
wt = fltarr(mmax+1,nz)  ; wt[0,*] is the sector-motion weight

; optional weights
if (n_elements(ufunc) ge 0) then ufunc = fltarr(mmax+1,nz)
if (n_elements(wtpup) ge 0) then   wtpup = fltarr(mmax+1,nz)

; finite-exposure averaging
if keyword_set(blur) then tfilt= f1(r*fstep*blur) else tfilt=1.

; Phase spectrum for Jturb=1
spturb = (r*fstep)^(-11./3.)*fstep^2*(9.62/2/!pi)*lambda^(-2)*tfilt
spturb[ngrid,ngrid] = 0.
; amplitudefluct. at phase screen =0, only phase fluct.
flux = total(imh*filtap) ; used for the normalization! 
; for convolution calculation
utmp = shift(fft(shift(conj(uampl),ngrid,ngrid)),ngrid,ngrid) 
x1d  = findgen(ngrid) ; 1D coordinate vector for spectral response

; Pre-compute 1-D Fresnel factors
frecos = fltarr(ngrid,nz)
fresin = fltarr(ngrid,nz) 
for iz=0,nz-1 do begin
   zdist = z[iz]
  if (nwav gt 1) then begin ; full spectrum is given
   spcos = fltarr(ngrid) & spsin = fltarr(ngrid)
   damp1 =  exp(- ( 0.5*(x1d*fstep)^2*wavstep*zdist)^2 ) ; remove ringing
   for j=0,nwav-1 do begin
    arg = !pi*wav[j]*zdist*x1d^2*fstep^2
    a = spnorm[j]*lambda/wav[j]
    spcos += a*cos(arg)
    spsin += a*sin(arg)
   endfor  
   frecos[*,iz] = spcos/spcos[0]*damp1
   fresin[*,iz] = spsin/spcos[0]*damp1

  endif else begin ; monochromatic
;    const = lambda*zdist*fstep^2
;    damp = exp(-0.5*1.78*const^2*x1d^4*(dlam/lambda)^2)  ; bandwidth damping
    frecos[*,iz] = cos(!pi*lambda*zdist*x1d^2*fstep^2) ; *damp
    fresin[*,iz] = sin(!pi*lambda*zdist*x1d^2*fstep^2)  ; *damp
  endelse ; dlam
endfor ; iz Fresnel

; stop

;---------- Main loops over m and z
for m=0,mmax do begin  ; ------ loop over m

 if (m eq 0) then begin ; radial masks
   cwt = rwt
   swt = 0
   pixfact = 1.
 endif else begin; angular masks 
   cwt = cos(phi*m)*filtap
   swt = sin(phi*m)*filtap
   if keyword_set(pixel) then begin ;pixel smoothing correction
     arg = (pixel/asperpix)/(1.5*ringradpix2/(m > 1)) ; ratio of pixel to mask period
     pixfact = (sin(arg)/arg) ; weight multiplier to account for pixel averaging
    endif else pixfact = 1.    
 endelse  ; m>0
   
; --- Compute the phase and amplitude filters -----
; response to the cosine mask
result1 = uampl*shift(fft(shift(utmp*cwt,ngrid,ngrid),/inverse),ngrid,ngrid)
; if cwt=1,  result1 is the pupil transmission U*U^* function, 1 inside pupil

; --- The fphase and fampl express relative fluctuations  --------
; Independent of the u-normalization, surface = total(abs(u)^2)
if (m gt 0) then fact=1./flux else fact=1./rflux  ; normalization
  fphase1 = -imaginary(result1)*fact
  fampl1 = real_part(result1)*fact   
  
; Sine response, similar to cosine
if (m gt 0) then begin 
   result2 = uampl*shift(fft(shift(utmp*swt,ngrid,ngrid),/inverse),ngrid,ngrid)
   fphase2 = -imaginary(result2)*fact
   fampl2 = real_part(result2)*fact   
endif ; m>0
   
if (n_elements(wtpup) ge 0) then begin ; prepare pupil-plane filter
  pfilt = shift(abs(fft(ur*cos(m*phi), /inverse))^2,ngrid, ngrid) + $
         shift(abs(fft(ur*sin(m*phi), /inverse))^2,ngrid, ngrid)
  pfilt = pfilt/surface^2
endif

; ---------------- Propagation  ---------------------
for iz=0,nz-1 do begin ; loop over Z
  zdist = z[iz]
  ctmp = frecos[r+0.5,iz] ; 2D Fresnel-factor arrays
  stmp = fresin[r+0.5,iz] 
;  ctmp = frecos[r,iz] ; 2D Fresnel-factor arrays
;  stmp = fresin[r,iz] 

  tmp1 = shift(fft(shift(fphase1,ngrid,ngrid),/inverse),ngrid, ngrid)*ctmp +  $
 shift(fft(shift(fampl1,ngrid,ngrid),/inverse),ngrid,ngrid)*stmp

  if (m gt 0) then $
       tmp2 =  shift(fft(shift(fphase2,ngrid,ngrid),/inverse),ngrid, ngrid)*ctmp +  $
        shift(fft(shift(fampl2,ngrid,ngrid),/inverse),ngrid, ngrid)*stmp $
  else tmp2 = 0.
  
  pfilter = abs(tmp1^2) + abs(tmp2^2) ; 

;  stop
  wt[m,iz] = total(pfilter*spturb)*pixfact

  
 if (n_elements(wtpup) ge 0) and (m gt 0) then  wtpup[m,iz] = total(pfilt*spturb*stmp^2)
 if (n_elements(ufunc) ge 0) and (m gt 0) then ufunc[m,iz] = !pi^2 *total(pfilter*spturb*(r*fstep)^2)
 
;  plot, (spturb*pfilter)[ngrid:ngrid+100,ngrid]
;  stop  ; testing

endfor ; iz
endfor ; m

;stop
return, wt

end

;-------------------------------------------------------
;------ end of the program ----------------------------  
