; Process cube of ring-like images. see cubespec.pro
; input: cube, par
; output: coef[2Nsect+2MMax+2, nz] matrix of frame coefficients
; impar - image parameters
;--------------------------------------------------------
pro cubecoef, cube, par, coef, impar, imav=imav, display=display, flat=flat 

nstart = 50. ; initial average
leak = 1.0   ; for running average of centroid 

s = size(cube)
nx = s[1] & ny = s[2] & nz = s[3]
if (ny ne nx) then begin
  print, 'Non-square frames! Returning'
  stop
  return
endif

if (nz lt nstart) then begin
  print, nz,' frames in the cube, too few'
  return
endif

; Vectors of radius and coordinates
xx = (findgen(nx) - nx/2) # replicate(1,ny)
yy =  transpose(xx)
r = sqrt(xx^2 + yy^2)
phi = atan(yy, xx) ; + !pi ; 0-2pi range
phi[0:nx/2,nx/2] = -!pi ; was pi and created empty ray
x1d = findgen(nx)  ; 1-D coordinate for interpolation

;prepare the filters and output arrays
m = par.mmax
nsect = par.nsect
rwt = fltarr(nx,nx,nsect)  ; for radius calculation
fwt = fltarr(nx,nx,nsect)  ; fluxes in the sectors
sect = 2.*!pi/nsect ; sector with, radian
for j=0,nsect-1 do begin  ; un-masked sectors for initial estimate
;   sector = (phi ge sect*j) and (phi lt sect*(j+1))
   sector = (phi ge (sect*(j- nsect/2))) and (phi lt (sect*(j+1-nsect/2)))
   fwt[*,*,j] = sector
   rwt[*,*,j] = sector*r 
endfor ; j sectors
;phisect = sect*(findgen(nsect)  +0.5) ; sector angles
phisect = sect*(findgen(nsect) - nsect/2 +0.5) ; sector angles

xsect = cos(phisect) & ysect = sin(phisect)

; Find initial ring parameters
imav = total(float(cube[*,*,0:nstart-1]),3)/nstart ; average nstart frames
backgr = median( [imav[0,*],imav[nx-1,*]]) ; left and right columns
tmp = imav - backgr
itot = total(tmp)
xc = total(tmp*xx)/itot
yc = total(tmp*yy)/itot
imgcent = shift(tmp, -xc, -yc) ; crude shift

radii = fltarr(nsect)
for j=0,nsect-1 do radii[j] = total(imgcent*rwt[*,*,j])/total(imgcent*fwt[*,*,j])
radpix = total(radii)/nsect  ; average radius
dx1 = total(radii*xsect)/nsect*2.3 ; x-shift, empirical 2.3
dy1 = total(radii*ysect)/nsect*2.3  ; y-shift
xc = xc+dx1  & yc = yc+dy1

; FFT shift
arg = 2.*!pi*(dx1*xx + dy1*yy)/nx
imgcent = float( fft(fft(imgcent)*shift(complex(cos(arg), sin(arg)), nx/2, nx/2), /inverse))

;imgcent = bilinear(imgcent, x1d+dx1, x1d+dy1) ; better-centered
;imgcent = shift(tmp, -xc, -yc) ; better-centered image
; centering control:
; plot, xx[*,0], imgcent[*,nx/2]
; oplot, -xx[*,0],  imgcent[*,nx/2], li=2
;stop

; diffraction-limited ring half-width, pixels
rwidthmin = par.lambda/par.d/(1. - par.eps)*2.*206265./par.pixel 
tmp = (imgcent - 0.1*max(imgcent)) > 0  ; thresholded image
radvar = total(tmp*(r - radpix)^2)/total(tmp)
rwidth = sqrt(radvar)*2.35
; rwidth = rwidth ; > rwidthmin  ; no less than theoretical

print, 'Ring rad, width, minwidth [pix]: ', radpix, rwidth, rwidthmin, format='(A,3F8.2)'

if (radpix gt nx/4) then begin   ; empty cube?
    impar = {backgr:backgr,flux:-1.,rad:radpix}
    return
endif

; recompute background
out = where(r gt 1.5*radpix)
backgr += median(imgcent[out])
; stop

; General  matrix of masks
;ringmask = exp(- (r - radpix)^2/(2.*(par.drad*rwidth^2)))  ; Gaussian mask
drhopix = par.drad*rwidth     ; actual
;drhopix = par.drad*rwidthmin  ; theoretical
ringmask = (r ge radpix-drhopix) and (r le radpix+drhopix) ; Circular, use only for display
inring = where((r ge radpix-drhopix) and (r le radpix+drhopix), nring) ; pixels inside radial mask 

ncoef = 2*nsect + 2*(m+1)
maskmat = fltarr(ncoef, nring) 
;maskmat = fltarr(ncoef, nx^2) 

; Mask sectors radially, insert into the matrix
for j=0,nsect-1 do begin
  maskmat[j,*] = (rwt[*,*,j])[inring]
  maskmat[j+nsect,*] = (fwt[*,*,j])[inring]
endfor ;j  

cwt = fltarr(nx,nx,m+1) ; cosine weight
swt = fltarr(nx,nx,m+1) ; sine weight
for j=0,m do begin 
  tmp =  cos(phi*j)*ringmask
  if (j gt 0) then cwt[*,*,j] = tmp  - total(tmp)/nx/nx else cwt[*,*,j] = tmp 
  tmp = sin(phi*j)*ringmask
  swt[*,*,j] = tmp  - total(tmp)/nx/nx
  maskmat[2*nsect+j,*] =  (cwt[*,*,j])[inring]
  maskmat[2*nsect+m+1+j,*] = (swt[*,*,j])[inring]
endfor ; j angular

; control of sector definition
if keyword_set(display) then begin
  imax = max(imgcent[*,nx/2])
  plot, xx[*,0], imgcent[*,nx/2] > 0, xs=1
  oplot,  -xx[*,0], imgcent[*,nx/2], li=2
  oplot, -radpix+rwidth*[-1,1], imax/2*[1,1], thick=4 
  oplot, +radpix+rwidth*[-1,1], imax/2*[1,1], thick=4 
;  tmp = fltarr(nx,nx)
;  for j=0,nsect-1 do tmp += fwt[*,*,j]*(j+1)
;  tvscl, tmp  ; control
  print, '.C to continue with cube processing'
  stop
endif ; display  

coef = fltarr(ncoef, nz) ; all coefficients
xcent = fltarr(nz) & ycent = fltarr(nz) & rad = fltarr(nz)
; dxtrack = fltarr(nz) & dytrack = fltarr(nz) ; test arrays

x0 = xc & y0 = yc & rad0 = radpix ; running centroids
print, 'Processing the cube...'
imav = fltarr(nx,nx) ; re-centered image

; ------- Main loop over the cube ------------------
for i=0,nz-1 do begin
  tmp = float(cube[*,*,i]) - backgr
  if keyword_set(flat) then tmp = tmp/flat ; flat-field division  

if par.interpol then begin ; sub-pixel shift
; bilinear shift
;  ix = floor(x0) & iy = floor(y0)
;  tmp = shift(tmp, nx-ix, nx-iy) ; re-centered frame
;  tmp = bilinear(tmp, x1d+x0-ix, x1d+y0-iy) ; interpolated
; FFT shift
  arg = 2.*!pi*(x0*xx + y0*yy)/nx
  tmp = float( fft(fft(tmp)*shift(complex(cos(arg), sin(arg)), nx/2, nx/2), /inverse)) 
;  tmp = fftshift(tmp, -x0, -y0)    ; FFT shift
endif else   tmp = shift(tmp, -x0, -y0) ; integer shift

; centering control
; plot, xx[*,0], tmp[*,nx/2]
; oplot, -xx[*,0],  tmp[*,nx/2], li=2
    
  imav += tmp
  c = maskmat # tmp[inring]
  c[0:nsect-1] = c[0:nsect-1]/c[nsect:2*nsect-1] ; normalize radius
  coef[*,i] = c ; save
  
  ; update radius and centroid
  radii = c[0:nsect-1] 
  dr = total(radii)/nsect
  dx = total(radii*xsect)/nsect*2.3
  dy = total(radii*ysect)/nsect*2.3

;  dxtrack[i] = dx & dytrack[i] = dy ; save corrections  
;  if (i eq 0) then stop, 'i=0!!!'  ; DEBUG centroids
      
  x0 = x0 + leak*dx
  y0 = y0 + leak*dy
  rad0 = rad0*(1-leak) + dr*leak
  xcent[i] = x0 & ycent[i] = y0 & rad[i] = dr

  ; control plot
  if (keyword_set(display) and (i mod 50 eq 0)) then begin
    tmp1 = tmp/max(tmp)
    tmp1[radii*xsect+nx/2, radii*ysect+nx/2] = -0.5
    tmp1[nx/2, nx/2] = 1.
    tvscl, congrid(tmp1,256,256)
;    print, 'x0,y0: ',x0,y0,dx,dy, format='(A,2F7.2, " dx,dy: ", 2F6.2)'
;    stop, 'i = '+string(i)
  endif ; display

endfor
; print, 'loop done'
imav = imav/nz

flux = mean(coef[2*nsect,*]) ; m=0 coefficient in ADU
coef[2*nsect:ncoef-1,*] *= 1./flux ; norm. by mean flux
fluxvar = sqrt(variance(coef[2*nsect,*])) ; relative flux variance
rad =  mean(rad) ; mean radius, pixels
xc = mean(xcent) & yc = mean(ycent)
xcvar = sqrt(variance(xcent)) & ycvar = sqrt(variance(ycent))
; coma quantification
cm = mean(coef[2*nsect+1,*])
sm = mean(coef[2*nsect+m+2,*])
coma = sqrt(cm^2 + sm^2)
angle = !radeg*atan(sm,cm)

; ring contrast per sector
contrast = fltarr(nsect)
for j=0,nsect-1 do begin
  tmp = imav*fwt[*,*,j]
  contrast[j] = max(tmp)/total(tmp)
endfor

; noise calculation
tmp  = imav/total(imav)
t0 = total(tmp*ringmask)  ; ringmask efficiency 
noise1 = total(ringmask^2*tmp)/t0 ; photon-noise factor angular
noise2 = total(ringmask^2)/t0      ; readout noise angular

;tmp2 =  ringmask*r ; radius weight
tmp2 =  ringmask*(r-radpix) ; radius weight

;  t1 = (total(tmp2*tmp))^2 
noise1r = total( tmp*tmp2^2);  /t0 ; radius photon noise
noise2r =  total(tmp2^2) ;   /t0       ; radius readout noise
noisepar= [noise1, noise2, noise1r, noise2r]

; stop
impar = {backgr:backgr,flux:flux,fluxvar:fluxvar,rad:rad,rwidth:rwidth, $
       xc:xc,yc:yc,xcvar:xcvar,ycvar:ycvar,coma:coma,angle:angle, $
       contrast:mean(contrast),noisepar:noisepar}

;stop

; tvscl, congrid(imav,256,256)

if keyword_set(display) then begin

  plot, xcent, ycent, psym=1, /iso
; plot, xcent*par.pixel, ycent*par.pixel, psym=1, /iso, xtitle='X [arcsec]', ytitle='Y [arcsec]'
  ; oplot, ycent, li=2
  stop, 'DONE!!! .C to exit cubecoef'
     print, 'Exiting cubecoef'   
endif ; display

end
