; May 13-14, 2021: profile restoration and time-constant measurement. Uses aweight3.pro
; 
;------------------------------------------------------------------------------
PRO   svd_invert, matrix, inv_matrix, threshold
; Returns the SVD-inverted matrix 
svdc, matrix, ws, u, v , /double
ww = max(ws) & n = n_elements(ws)
invw = identity(n) & ncount = 0
FOR i=0, n-1 DO BEGIN 
  IF ws(i) LT ww*threshold THEN BEGIN 
     invw(i,i) = 0.
     ncount = ncount+1
  ENDIF  ELSE invw(i,i)=1./ws(i)
ENDFOR
print, ncount, ' singular values rejected in inversion'
inv_matrix =  (v ##  invw) ## transpose(u)
END 
;------------------------------------------------------------------------------
; Find U-coefficients. mm is the set of used terms, z is the z-grid
function getucoef, ufunc, z, mm
  uf = ufunc[mm,*] 
  nm = n_elements(mm) & nz = n_elements(z)
  ; remove lower alt from zgrid
  z1 = z[2:nz-1] & nz = nz-2

;  wz = z1*1e-3 + 0.5 ; distance weight
   wz = replicate(1., nz)

  amat = fltarr(nz,nm)
  for i=0,nm-1 do amat[*,i] = uf[i,*]*wz
  aa = transpose(amat) # amat   ; system matrix
  b = transpose(amat)# wz       ; right-hand part
  svd_invert, aa, aainv, 1e-4
  ucoef = reform(aainv ## b, nm)
;  plot, z1*1e-3, uf ## ucoef, xtitle='Z [km]', ytitle='U-response', xs=1, xr=[0.25,32], /xlog
;  print, '.C to continue'
;  stop
  return, ucoef
end
;------------------------------------------------------------------------------
; define the weights.
; the m=0 weight is for sector motion variance in lam/D units.
; Compute weights for B-V of 0 and 1 and approximate then linearly
; ampl is the coma amplitude
pro getweight5, par, ampl=ampl

mmax = 20  ; fixed parameters
nz = 16
z = [0, 1000.*2.^(0.5*findgen(nz-1)-2)] ; altitude grid 0, 0.25-32km

mm =  [1,3,6,7,8,9] ; selected frequencies for wind measurement
nm = n_elements(mm) ; number of angular frequencies

; read the spectral response (2 columns lamnda[nm], response)
s = findfile(par.qedata, count=count)
if (count eq 0) then begin
   print, 'Not found the spectral response '+par.qedata
   return
endif   
tmp = import(par.qedata, 2) 
q = tmp[*,1] & good = where(q gt 0)
q = q[good] & w= tmp[good,0] ; use only non-zero-response wavelengths
lam = w *1e-9  ; wavelength in meters

print, 'Computing weights for B-V of 0 and 1'
tcol = [10213., 3938.] ; color temperature for B-V of 0 and 1
spec0 = q*blackbody(lam,tcol[0])*(w/w[0]) ; response in photons
spec0 = spec0/total(spec0)
spec1 = q*blackbody(lam,tcol[1])*(w/w[0]) ; same for B-V=1
spec1 = spec1/total(spec1)
lameff = [total(spec0*lam), total(spec1*lam)]  ; effective wavelength
print, 'Effective wavelengths: ',lameff*1e9

if keyword_set(ampl) then begin ; Coma is included!
   print, 'Coma aberration is included! Ampl ', ampl
   a7 = ampl/0.057  ; coma ampl, radian
   zn=[3,7] & zrad=a7*[-4,1.]  ; coma parameters
   print, 'Computing weights for B-V=0...'
   wt0 = aweight3(z,mmax,par.d,par.eps,par.pdist,lam,spec0,pixel=par.pixel, ufunc=ufunc0,zn=zn,zrad=zrad)
   print, 'Computing weights for B-V=1...'
   wt1 = aweight3(z,mmax,par.d,par.eps,par.pdist,lam,spec1,pixel=par.pixel, ufunc=ufunc1,zn=zn,zrad=zrad)

endif else begin   
						   
print, 'Computing weights for B-V=0...'
wt0 = aweight3(z,mmax,par.d,par.eps,par.pdist,lam,spec0,pixel=par.pixel, ufunc=ufunc0)
print, 'Computing weights for B-V=1...'
wt1 = aweight3(z,mmax,par.d,par.eps,par.pdist,lam,spec1,pixel=par.pixel, ufunc=ufunc1)

endelse ; coma

wtslope = wt1 - wt0

plot, wt0[1,*], /ylog, ytitle='Weight [m!U-1/3!N]'
for j=2, 10 do oplot,  wt0[j,*]
for j=1, 10 do oplot,  wt1[j,*], color='0000AA'x

ucoef0 = getucoef(ufunc0,z,mm)
ucoef1 = getucoef(ufunc1,z,mm)  ; slope ~0.1
ucoefslope = ucoef1 - ucoef0

; for saturation correction: use m=1...5 power terms
if (par.zmat gt 0) then restore, 'zmat.idl' else zmat = fltarr(20,5)  ; z
; zmat = import('zmat.txt',5) ; also works

; relation between ring radius and propagation distance
rad = aweight3([0],1,par.d,par.eps,par.pdist,lameff[0],[1],pixel=par.pixel, /ringrad)
hslope = par.pdist*rad  ; m*pixel 

save, fi=par.weightfile, z, wt0, wtslope, ucoef0, ucoefslope, mm, lameff, zmat, hslope
print, 'DONE! Weights saved in '+par.weightfile
;stop
end
;------------------------------------------------------------------------------
; Study ring radius vs. conjugation
pro conjug, par
m=5
h0 = par.pdist*[0.8, 0.9, 1., 1.5, 2]

rad = fltarr(m)
z = [0] & lam = 0.6e-6 & spec = [1] 
for i=0,m-1 do rad[i] =  aweight3(z,1,par.d,par.eps,h0[i],lam,spec,pixel=par.pixel, /ringrad)

xx = [0, 1./h0[0]]
slope = rad[1]*h0[1]
print, 'Slope [pix.m]: ', slope

hh = interpol(h0, rad, par.ringradpix)
print, 'Nominal conjugation: ', hh

plot, 1/h0, rad, xr=xx, xtitle='1/H', ytitle='Ring rad [pix]'
oplot, 1./h0, rad, psym=6
oplot, xx, xx*slope, li=2

stop

plot, rad, h0, xtitle='Ring rad [pix]', ytitle='H [m]'

end
;------------------------------------------------------------------------------
; profile restoration. Uses getprof.pro and angular power spectrum POW,
; noise-subtracted at input!
; wt is projected on the Z0 grid, corrected for zenith dist. and B-V color
; ucoef is also B-V-corrected
pro profrest5, pow, cov, prof, erms, wind, wt, z0, ucoef, mm, zmat

mmax = 17 ; max number of used terms

rho = cov/pow ; discard fast frequencies
for i=1,mmax-1 do  if (rho[i] lt 0.) then break  ; check for negative covariances
mmax = i > 9    
 print, 'Max. frequency ',mmax

pow1 = pow[1:mmax] ; remove the first element
cov1 = cov[1:mmax]
totvar = total(pow)
nz = n_elements(z0)

rho = cov1/pow1 ; correlation coefficient
powcorr = pow1/(0.8 + 0.2*rho)
;  powcorr = pow1 ; disable exp-time corr.
powcorr *=  1./(1. + zmat # powcorr[0:4])  ;  saturation correction

; debug
; print, powcorr, format='(20E9.2)'
; stop

;epow =pow1/sqrt(200.)/sqrt(findgen(mmax)+1.) ; notional error, m-weighted
epow =pow1/sqrt(200.) ; notional error

;epow = replicate(1.,mmax) ; unit weight

;nz = n_elements(z0)
getprof, z0, wt[1:mmax,*], powcorr, epow, prof, chi2  ; compile nnls first!

powmod = wt[1:mmax,*] # prof
e = 1. - powmod/powcorr ; relative error
erms = sqrt(total(e^2)/mmax)
print, 'CHI2, Rel. rms to model: ', CHI2, erms, format='(A,F6.2,F8.4)'

!p.charsize = 1.5
arg = findgen(mmax)+1
plot, arg, powcorr, /ylog, psym=6, xtitle='m', ytitle='Power' , xs=1   
oplot, arg, powmod, li=2

; wind measurement
texp = 1e-3 ; hard-coded
delta = (pow - cov)[mm]*texp^(-2) > 0.
v2mom = total(delta*ucoef) > 0.
; jtot = total(prof)
jtot = total(prof[2:nz-1])

v2 = sqrt(v2mom/jtot)
; print, 'Measured wind speed [m/s]: ', v2
wind=v2

;stop
end
;------------------------------------------------------------------------------
; test profile restoration for one file
; bv is the B-V color, zen is zenith distance in degrees
pro testprof, par, el ; , num, bv, zen

;  restore, par.datfile ; dat
  restore, par.weightfile  ; weights, z, ucoef etc. from getweight
;  restore, 'wt5.idl'  ; weights, z, ucoef etc. from getweight
;  el = dat[num] ; selected element of the data structure
   bv = el.bv
   zen = el.zen
   
  nz = 8
  z0 = [0, 1000.*2.^(findgen(nz-1)-2)]  ; 0, 0.25km... 16km standard

;  nz = 7
;  z0 = [1000.*2.^(findgen(nz)-2)]  ; 0.25km... 16km without ground

; remove 250-m layer
;  nz = 7
;  z0 = [0, 1000.*2.^(findgen(nz-1)-1)]  ; 0, 0.25km... 16km standard
; add extra layer
;  nz = 9
; z0 = [50., 1000.*2.^(findgen(nz-1)-3)]  ; 0, 0.125km... 16km 
; More layers: 0, 125,250,353,... 16km
;  nz = 16
;  z0 = [0, 1000.*2.^( (findgen(nz-1)-6)/2.)]  ; 0, 0.25km... 16km standard

  secz = 1./cos(zen/!radeg) ; sec z
;  print, 'Restore data record #',num
  
; Interpolate weight to the stretched grid to account for propagation
; do not correct for sec(z) yet
  m = (size(wt0))[1]  ; number of weights
  wt = fltarr(m,nz)
  for i=0,m-1 do wt[i,*] = interpol(wt0[i,*] + bv*wtslope[i,*], z, z0*secz)  ; adjusted WF
  ucoef = ucoef0 + bv*ucoefslope

  pow = (el.pow - el.anoise) > 0.
  cov = el.cov
  totvar = total(pow)

 ; stop, 'Debug in testprof' ; debug

  profrest5, pow, cov, prof, erms, wind, wt, z0, ucoef, mm, zmat
  prof = prof/secz  ; correct for zenith dist.
  
jtot = total(prof[0:nz-1])
see = (jtot/6.826E-13)^0.6
jfree =  total(prof[3:nz-1])
fsee = (jfree/6.8E-13)^0.6

print, 'See, fsee:    ', see, fsee, format='(A,2F6.2)'
; print, 'Old   fsee:   ', el.see, el.fsee, format='(A,2F6.2)'
print, 'Scint: ', totvar, format='(A,F6.3, E10.2)'

;idx = [0,2,3,4,5,6,7,8] ; index into profile saved in data, old
print, 'Z: ', z0*1e-3, format='(A5,20F8.2)'
print, 'J: ',prof*1e13, format='(A5,20F8.2)'
; print, 'J1:', el.prof[idx]*1e13, format='(A,20F8.2)'
;print, 'Wind(new,old): ',wind, el.wind
print, 'Wind: ',wind

; Now  the sector variance
wtsect = total(wt[0,*]*prof)/total(prof) ; weighted sector weight
lam0 = lameff[0] + bv*(lameff[1] - lameff[0]) ; effective wavelength for sector weight
lamd = lam0/par.d*206265  ; lam/D in arcsec
radvar =   el.rrms^2 - el.rnoise1 ; square pixel 
rvarnorm = radvar*(par.pixel/lamd)^2 ; variance in lam/D units
jtot2 = rvarnorm/wtsect/4./secz ; turbulence integral at zenith
see2 =  (jtot2/6.826E-13)^0.6 ; seeing in arcsec
;see2 =  see2/(1. - 0.78*totvar)
 see2 =  see2/(1. - 0.40*totvar)

;print, 'Sect. seeing (new,old): ', see2/(1. - 0.78*totvar), el.see2, format='(A,3F6.2)'
print, 'Sect. seeing: ', see2, format='(A,3F6.2)'

;  stop
end
;------------------------------------------------------------------------------
; Restore all profiles for a given night. Use allprof5 to create the data stucture,
; and to compute zenith distance and BV color.
pro allprof5, par

restore, par.weightfile
restore, par.datfile ; dat
n = n_elements(dat)
nz = 8
z0 = [0, 1000.*2.^(findgen(nz-1)-2)]  ; 0, 0.25km... 16km
m = (size(wt0))[1]  ; number of weights
wt = fltarr(m,nz)   ; interpolated weight

for i=0,n-1 do begin
  print, 'i= ',i
  el = dat[i]
  pow = (el.pow - el.anoise) > 0.
  cov = el.cov
  totvar = total(pow)
  bv = el.bv
  secz = 1./cos(el.zen/!radeg) ; sec z
  for j=0,m-1 do wt[j,*] = interpol(wt0[j,*] + bv*wtslope[j,*], z, z0*secz)  ; adjusted WF
  ucoef = ucoef0 + bv*ucoefslope
  
  profrest5, pow, cov, prof, erms, wind, wt, z0, ucoef, mm, zmat
  prof = prof/secz  ; at zenith

  ;  sector variance
  wtsect = total(wt[0,*]*prof)/total(prof) ; weighted sector weight
  lam0 = lameff[0] + bv*(lameff[1] - lameff[0]) ; effective wavelength for sector weight
  lamd = lam0/par.d*206265  ; lam/D in arcsec
  radvar =   el.rrms^2 - el.rnoise1 ; square pixel 
  rvarnorm = radvar*(par.pixel/lamd)^2 ; variance in lam/D units
  jtot2 = rvarnorm/wtsect/4./secz ; turbulence integral at zenith
  see2 =  (jtot2/6.826E-13)^0.6 ; seeing in arcsec
  see2 = see2/(1. - 0.78*totvar) ; correct for saturation
;  see2 = see2/(1. - 0.4*totvar) ; correct for saturation, but not so much!

  dat[i].prof = prof 
  jtot = total(prof)
  see = (jtot/6.826E-13)^0.6
  jfree =  total(prof[2:nz-1]) ; 0.5km and above
  fsee = (jfree/6.826E-13)^0.6
  dat[i].see = see
  dat[i].fsee = fsee
  dat[i].rmserr =  erms
  dat[i].scint = total(pow)
  dat[i].wind = wind
  dat[i].see2 = see2
  ; stop
endfor ; i
print, 'Done!'
	   ;stop
save, fi=par.datfile, dat
end
;------------------------------------------------------------------------------
;------------------------------------------------------------------------------
