; Monte-Carlo simulation of defocused images with seeing
; Use pre-computed complex amplitude (simatm.idl) saved in the atm.idl
; compile ZTOOLS first if aberration is applied
; Example: simatm, 512, 0.005, 0.6e-6, 0.1, 0.0, 0, 1e4
; Aug 25, 2020; Feb 21, 2023
;-------------INPUT PARAMETERS: ------------------------
; ZRAD, ZN - vectors of Zernike(Noll) aberrations in radians and their numbers
; jitter - amplitude of circular motion in pixels to simulate tracking
;---------------OUTPUT PARAMETERS ----------------------
; angz - angular spectrum, cube - image cube 
;-------------------------------------------------------
;-------------------------------------------------------
pro ringsim, par, cube, zn=zn, zrad=zrad, starmag=starmag, ron=ron, wind=wind, blur=blur, debug=debug, jitter=jitter

; -----  Hard-coded parameters -----------------
if not keyword_set(starmag) then starmag = 0.
if not keyword_set(ron) then  ron = 1.
    
D = par.d
EPS = par.eps
pdist = par.pdist

; hard-coded parameters    
texp = 0.001
tacc = 2.  ; accumulation time, seconds
; ksamp = 4  ; oversampling of atm. screen, defines pixel size and field
; npixperpix = 4    ; CCD binning factor
if not keyword_set(wind) then wind = 10.

;--------------------------------------------

restore, 'atm.idl'
; restores u1,ngrid,pixel,lambda,see,r0,fhigh,zlow,zhigh

if (lambda ne par.lambda) then begin
  print, 'Mismatch in simulated wavelength, use ',par.lambda
  return
endif  

if keyword_set(blur) then begin
  nblur = fix(wind*texp/pixel + 0.5)
  print, 'Averaging atmospheric screens, N=',nblur
  if (nblur gt 1) then begin
     tmp = u1
     for k=1,nblur-1 do tmp+=shift(u1,k,0)
     u1 = tmp/nblur
  endif ; nblur>1
endif ; blur



  if n_elements(zlow) eq 0 then zlow=zhigh ; undefined zlow sometimes
  see = 0.98*lambda/r0*206265.
  print, 'Seeing ',see,' arcsec, layers at ',zlow,zhigh,' high fraction ',fhigh
  print, 'Total screen size,m, and pixel: ', 2*ngrid*pixel, pixel  
;  r0 = 0.98*lambda/(see/206265.) ; Fried parameter, m
 tint = r0^(-5./3.)/0.423*(0.5*lambda/!pi)^(2) ; total turbulence integral, m^1/3
 print, 'Input r0, J: ', r0,tint

 
; Apertures move over the screen mostly in x-direction, but slide
; in y-direction by SLIDE meters par grid length

size = 2*ngrid*pixel
slide = 0.205
alpha = slide/size  ; tangent of the slide angle

niter = floor(tacc/texp+1)  ; number of steps
jstep = floor(wind*texp/pixel + 0.5)  ; integer pixel shift
if (jstep lt 1) then jstep=1
windef = jstep*pixel/texp
 print, 'Screen shift per exposure, pixels ',jstep 
 print, 'Effective wind speed, m/s ',windef

; ------- define grid size and over-ampling as in rmassnoise
nap = 256 ; size of aperture and image arrays in fine pixels
;nap = 512 ; size of aperture and image arrays in fine pixels, too much

d1 = lambda/(par.pixel)*206265. ; pupil scale matching pixels
npixperpix =  2^(floor(alog(1.5*d/d1)/alog(2.))+1) ; over-sampling

nscr = 2^(floor(alog(1.5*d/pixel)/alog(2)) +1) ; screen size in pixels
ksamp = nap/nscr > 1

print, 'Over-sampling factor: ', ksamp

;stop  ; test stop
						   
;print, 'Aperture grid, pixels: ', nap
;print, 'Re-sampled pixel size [m]: ', pixel/ksamp
asperpix = 206265.*lambda / (nap*pixel/ksamp)*npixperpix
nccd = nap/npixperpix ; number of CCD pixels
print, 'CCD size & pixel [arcsec]: ', nccd, asperpix
; stop

ringradpix = 0.85*d*(1 + eps)/(4.*pdist)*206265./asperpix ; 0.85 added
print, 'Nominal ring radius [pix, arcsec]:', ringradpix, ringradpix*asperpix

if (abs(asperpix/par.pixel -1) gt 0.05) then begin
   print, 'Discrepant pixel scale in par! ', par.pixel
   print, 'To match, use SIMATM with pixel of ', pixel*asperpix/par.pixel
   stop
endif
par.pixel = asperpix

if keyword_set(jitter) then begin ; circular image shifts
  x1d = findgen(nccd)
  omega = 3.3/niter*(2.*!pi) ; angular frequency
endif


; star photons per aperture: total QE*bandwidth = 0.26 micron
if (starmag ne 0.) then begin
  BW= 0.26
  starph = 1.e11*texp*BW*10^(-0.4*starmag)*!pi*(D/2.)^2*(1. - eps)^2
  print, 'Stellar photons: ',starph, ' Readout noise: ',ron
endif
 
; --- Prepare the  aperture mask ---
apert = fltarr(nap,nap)
r = shift(dist(nap),nap/2,nap/2) ;*pixel/ksamp ; radial distance, pixels
Radpix = d*0.5/pixel*ksamp 
inside = where((r le Radpix) and (r ge Radpix*eps), nn)
if (nn gt 0) then apert(inside) = 1.
; apert[0:nap/2, nap/2-10:nap/2+10] = 0. ; test only: mask a sector

; Add phase distortions: defocus and spherical. a4 negative for intra-focal.

a4 = d^2/(lambda*pdist) *(!pi/8/sqrt(3.)) ; Zernike in radians
a11 = -0.1*a4  ; matching spherical
tmp = a11*sqrt(5.)*(6.*(r/Radpix)^4 - 6.*(r/Radpix)^2)
tmp += a4*2.*sqrt(3.)*((r/Radpix)^2 - 0.5)
print, 'Nonimal a4,a11 [rad]: ',a4,a11

; Add Zernike aberrations if specified, use ztools. See sim
if keyword_set(zn) then begin
  x = (findgen(nap) - nap/2) # replicate(1., nap)
  theta = atan(transpose(x),x)
  ninside = n_elements(inside)
  zgrid = fltarr(ninside,2)
  zgrid[*,0] = r(inside)/Radpix
  zgrid[*,1] = theta(inside)
  ; -- use Ztools.pro here! ---------------------
  phase = zgrid[*,0]*0. ; empty array
  for k=0,n_elements(ZRAD)-1 do phase += ZRAD[k]*zernike_estim(zn[k], zgrid)
  tmp[inside] += phase
endif

fresnel = complex(cos(tmp), sin(tmp))*apert
; undistorted image in fine pixels
imh0 = abs(shift(fft(shift(fresnel,nap/2,nap/2),/inverse),nap/2, nap/2))^2
normconst = total(imh0)

rring = total(imh0*r)/normconst
print, 'True ring radius [pix,arcsec]: ', rring/npixperpix, rring*asperpix/npixperpix

; plot, imh1[nap/2,*], xs=1
; oplot, rring*[-1,1]+nap/2, 0.5*max(imh1)*[1,1], psym=6

if keyword_set(debug) then begin
 tvscl, imh0 
 stop, 'DEBUG! .C to continue'
endif 

 
cube = fltarr(nccd,nccd,niter) ; image cube
y = 0 & x= 0       ; initial coordinates of the screen


;---------  Main loop ----------------------
print, 'Computing ',niter, ' iterations, count to 20...'
; stop

ndispl = niter/20  ; display 20 pictures
count=niter
ix = 0L & y = 0. 

for i=0L,niter-1 do begin
  ; carve a piece of the large screen shifted by ix,iy
    ix += jstep   & y += jstep*alpha &  iy = floor(y)

  ; shift the input screen if we pass the center of the array
  if (ix gt ngrid) then begin
    u1 = shift(u1, -ngrid, 0)
    ix = ix - ngrid
  endif
  if (iy gt ngrid) then begin
    u1 = shift(u1, 0, -ngrid)
    y = y - ngrid
    iy = floor(y)
  endif

  uampl = u1(ix:ix+nscr-1, iy:iy+nscr-1) 
  if (ksamp gt 1) then uampl = congrid(uampl, nap,nap, /interp)

  ; propagate to the detector plane
;  tmp = shift(fft(shift(uampl*apert,nap/2,nap/2)),nap/2, nap/2)

  imh1 = abs(shift(fft(shift(fresnel*uampl,nap/2,nap/2),/inverse),nap/2, nap/2))^2
  impix = rebin(imh1,nccd,nccd)*npixperpix^2/normconst ; normalized

  if keyword_set(jitter) then begin ; shift the binned image
      xc = jitter*cos(i*omega) & yc = jitter*sin(i*omega)
      impix = bilinear(impix, x1d+xc, x1d+yc)
   endif ; jitter

; --- Photon and readout noise
if (starmag ne 0) then begin
  impix *= starph 
  impix = poidev(impix,seed=seed) +  randomn(seed,nccd,nccd)*ron
endif ; starmag

; stop

;  imh2 = abs(shift(fft(shift(conj(fresnel)*tmp,nap/2,nap/2),/inverse),nap/2, nap/2))^2
   imh2 = abs(uampl*apert)^2 ; pupil image

  cube[*,*,i] =  impix


; ----- display and save ---------------------------------------
if (i mod ndispl) eq 0 then begin ; display every 10-th image
 tvscl, impix
 tvscl, congrid(imh2,nccd,nccd), nccd+2,0

;  tvscl, congrid(impix,2*nap,2*nap)
;  tvscl, congrid(imh2,2*nap,2*nap), 2*nap+2,0

 print,i/ndispl,format='(I3,$)'
 if keyword_set(debug) then stop, 'DEBUG! .C to continue'
 ; stop
endif  ; niter

  endfor
; --------- end of main loop -----------------------------
print, ' Simulation done!'
; stop
;imav = total(cube,3)/niter
;tvscl, congrid(imav,2*nap,2*nap)

; stop

end
