; Simulate monochromatic data cube and restore turbulence profile
; compile nnls, profrest5 before running this code!
pro testsimul

seeing = 1.0 ; seeing in arcsec at 0.5 micron
zlow = 500.  ; lower layer altitude [m]
zhigh = 11000. ; higher layer altitude
highfrac = 0.1
starmag = 2.  ; star magnitude

par = readpar('sim1.par') ; parameters of simulated instrument

lambda = par.lambda
r0 = 0.98*lambda/seeing*206265.  ; correct if seeing is specified at lambda
;tint = r0^(-5./3.)/0.423*(0.5*lambda/!pi)^(2) ; total turbulence integral, m^1/3
tint = 6.823e-13*seeing^(5./3.)
r0 = 0.101/seeing*(lambda/0.5e-6)^1.2

print, 'Input r0, J: ', r0,tint

;stop
; simulate the complex-amplitude screen
simatm, 512, 0.0025, lambda, r0, highfrac, zlow, zhigh

ringsim, par, cube, wind=5, starmag=starmag, ron=par.ron  ; simulate cube

writefits, 'cube.fits', cube

cubecoef, cube, par, coef, impar  ;
;cubecoef, cube, par, coef, impar, imav=imav, /displ  ;
print, 'Star flux [el]: ', impar.flux

; for reference:
; impar = {backgr,flux,fluxvar,rad,rwidth,
;    xc,yc,xcvar,ycvar,coma,angle,contrast,noisepar[4])}

statmom,  coef, par, pow, rrms, rnoise, cov=cov

; Compute noise on angular coefficients
flux = impar.flux ; star flux, photons per frame
anoise = impar.noisepar[0]/flux + impar.noisepar[1]*(par.ron/flux)^2
flux1 = flux/par.nsect ; flux per sector
rnoise1 =  2.*(impar.noisepar[2]/flux1 + impar.noisepar[3]*(par.ron/flux1)^2/par.nsect)

; Plot angular power spectum, covariance, and noise
arg = findgen(par.mmax +1)
yr = [anoise, max(pow)]
plot, arg, pow, yr=yr, xtitle='m', ytitle='Power', /ylog
oplot, arg, cov, li=2
oplot, [1,par.mmax+1], anoise*[1,1], li=1

print, 'Press .C to restore turbulence profile'
stop

; Skip weight calculation if done already

; getweight5, par

print, 'Ready for profile restoration!'

restore, par.weightfile  ; z, wt0, wtslope, ucoef0, ucoefslope,mm,lameff, hslope 

nz = 8
z0 = [0, 1000.*2.^(findgen(nz-1)-2)]  ; 0, 0.25km... 16km standard layers

pow = (pow - anoise) > 0.
totvar = total(pow)

m = (size(wt0))[1]  ; number of m-terms in pre-computed weight
wt = fltarr(m,nz)

for i=0,m-1 do wt[i,*] = interpol(wt0[i,*], z, z0)  ; adjusted WF

; stop

profrest5, pow, cov, prof, erms, wind, wt, z0, ucoef0, mm, zmat

jtot = total(prof[0:nz-1])
see = (jtot/6.826E-13)^0.6
jfree =  total(prof[2:nz-1])
fsee = (jfree/6.8E-13)^0.6

print, 'See, fsee:    ', see, fsee, format='(A,2F6.2)'
print, 'Scint: ', totvar, format='(A,F6.3, E10.2)'

print, 'Z [km]:   ', z0*1e-3, format='(A12,10F7.2)'
print, 'J [1e-13]:',prof*1e13, format='(A12,10F7.2)'

print, 'Wind: ',wind, format='(A,F7.1)'

; Alternative seeing from sector-radius variance
wtsect = total(wt[0,*]*prof)/total(prof) ; weighted sector weight
lamd = lambda/par.d*206265  ; lam/D in arcsec
radvar =  rrms^2 - rnoise1 ; square pixels  variance, noise-corrected
rvarnorm = radvar*(par.pixel/lamd)^2 ; variance in lam/D units
jtot2 = rvarnorm/wtsect/4. ; turbulence integral at zenith
see2 =  (jtot2/6.826E-13)^0.6 ; seeing in arcsec
see2 =  see2/(1. - 0.40*totvar) ; saturation correction
print, 'Sector seeing: ', see2, format='(A,3F6.2)'

print, 'Input seeing and J: ', seeing, tint*1e13, format='(A,2F6.2)'
print, 'Altitudes: ', zlow, zhigh, format='(A,2F8.1)'

print, 'Simulated cube is processed!'
stop

end
